part of insync;

abstract class AbstractCollectionSorterComparing<V, C extends AbstractCollection<V>>
    extends Owner {
  Function _compare;

  C _source;
  C get source => _source;

  AbstractList<V> _target;
  AbstractList<V> get target => _target;

  AbstractCollectionSorterComparing(this._source, {AbstractList<V> target,
    int compare(V x, V y)}) :
      _compare = compare {
    assert(_source != null);
    assert(_compare != null);
    _target = (target != null) ? target : _source.createEmptyList()..ownBy(this);
    _splice([], _source.asList());
  }

  void destroy() {
    _splice(_source.asList(), []);
    super.destroy();
  }

  void resort() {
    _target.sortComparing(_compare);
  }

  void _splice(List<V> removedValues, List<V> addedValues) {
    List<V> removedValuesSorted = new SimpleList<V>.adapter(removedValues).toSortedComparing(_compare);
    List<V> addedValuesSorted = new SimpleList<V>.adapter(addedValues).toSortedComparing(_compare);
    removedValues = new List.filled(removedValues.length, null);
    addedValues = new List.filled(addedValues.length, null);
    int iRemoved = 0;
    int iAdded = 0;
    int jRemoved = 0;
    int jAdded = 0;
    // ignore out the items which are removed and added at the same time
    while ((iRemoved < removedValuesSorted.length) || (iAdded < addedValuesSorted.length)) {
      V removedValue = removedValuesSorted[iRemoved];
      V addedValue = addedValuesSorted[iAdded];
      int c = compareBool(removedValue == null, addedValue == null);
      if (c != 0) {
        c = _compare(removedValue, addedValue);
      }
      if (c < 0) {
        removedValues[jRemoved++] = removedValue;
        ++iRemoved;
      } else if (c > 0) {
        addedValues[jAdded++] = addedValue;
        ++iAdded;
      } else {
        ++iRemoved;
        ++iAdded;
      }
    }
    removedValues.removeRange(jRemoved, removedValues.length);
    addedValues.removeRange(jAdded, addedValues.length);

    int iTarget = 0;
    int iAdds = 0;
    int addShift = 0;
    List<IndexCount> removeParamsList = [];
    List<IndexValues<V>> addParamsList = [];
    IndexCount removeParams;
    for (int lTarget = _target.length; iTarget < lTarget; ++iTarget) {
      V value = _target[iTarget];
      int index = new ListUtil<V>().binarySearch(removedValues, (V v) => _compare(v, value) >= 0);
      if (removedValues[index] == value) {
        if (removeParams == null) {
          removeParams = new IndexCount(iTarget, 0);
          removeParamsList.add(removeParams);
        }
        ++removeParams.count;
        --addShift;
      } else {
        removeParams = null;
        IndexValues<V> addParams = new IndexValues<V>(iTarget + addShift, []);
        while ((iAdds < addedValues.length) && (_compare(addedValues[iAdds], value) < 0)) {
          addParams.values.add(addedValues[iAdds++]);
          ++addShift;
        }
        if (addParams.values.isNotEmpty) {
          addParamsList.add(addParams);
        }
      }
    }
    if (iAdds < addedValues.length) {
      addParamsList.add(new IndexValues<V>(iTarget + addShift, addedValues.sublist(iAdds)));
    }
    _target.trySplice(removeParamsList, addParamsList);
  }
}