part of insync;

abstract class AbstractMap<K, V> extends IndexedCollection<K, V> {
  Map<K, V> _base = new HashMap<K, V>();
  Map<K, V> _constantMap;

  AbstractMap<K, dynamic> createEmpty();

  AbstractMap() {
    _constantMap = new ConstantMap(_base);
  }

  AbstractMap.from(Map<K, V> values) {
    _constantMap = new ConstantMap(_base);
    new MapUtil<K, V>().trySetAll(_base, values);
  }

  AbstractMap.adapter(this._base) {
    _constantMap = new ConstantMap(_base);
  }

  Map<K, V> get base => _constantMap;
  int get length => _base.length;
  bool get isEmpty => _base.isEmpty;
  K get firstKey => isEmpty ? null : _base.keys.first;
  V operator [](K key) => _base[key];

  List<K> getKeys() {
    return _base.keys.toList();
  }

  bool every2(bool f(V value, K key)) {
    for (K key in _base.keys) {
      if (!f(_base[key], key)) {
        return false;
      }
    }
    return true;
  }

  Map<K, V> filter(bool f(V value)) {
    return filter2((V v, K k) => f(v));
  }

  SimpleMap<K, V> $filter(bool f(V value)) {
    return new SimpleMap.adapter(filter(f));
  }

  Map<K, V> filter2(bool f(V value, K key)) {
    Map<K, V> result = new HashMap();
    every2((V v, K k) { if (f(v, k) != false) { result[k] = v; }});
    return result;
  }

  SimpleMap<K, V> $filter2(bool f(V value, K key)) {
    return new SimpleMap.adapter(filter2(f));
  }

  Map<K, dynamic> map(dynamic f(V value)) {
    return map2((V v, K k) => f(v));
  }

  SimpleMap<K, dynamic> $map(dynamic f(V value)) {
    return new SimpleMap.adapter(map(f));
  }

  Map<K, dynamic> map2(dynamic f(V value, K key)) {
    Map<K, dynamic> result = new HashMap();
    every2((V v, K k) { result[k] = f(v, k); });
    return result;
  }

  SimpleMap<K, dynamic> $map2(dynamic f(V value, K key)) {
    return new SimpleMap.adapter(map2(f));
  }

  Map<K, V> asMap() {
    return base;
  }

  AbstractMap<K, V> $asMap() {
    return this;
  }

  Wrap<V> trySet(V value, K key) {
    MapSpliceResult<K, V> spliceResult = trySplice([], new MapUtil<K, V>().single(key, value));
    return (spliceResult != null) ? new Wrap(spliceResult.removedValues[key]) : null;
  }

  MapSpliceResult<K, V> setAll(Map<K, V> values) {
    MapSpliceResult<K, V> result = trySetAll(values);
    return (result != null) ? result : new MapSpliceResult();
  }

  MapSpliceResult<K, V> trySetAll(Map<K, V> values) {
    return trySplice([], values);
  }

  V setKey(K oldKey, K newKey) {
    trySetKey(oldKey, newKey);
    return _base[newKey];
  }

  V trySetKey(K oldKey, K newKey) {
    Map<K, K> keyMap = tryReindex(new MapUtil<K, K>().single(oldKey, newKey));
    return (keyMap != null) ? _base[newKey] : null;
  }

  V tryRemove(K key) {
    MapSpliceResult<K, V> result = trySplice([key], {});
    return (result != null) ? result.removedValues[key] : null;
  }

  Map<K, V> removeAll(List<K> keys) {
    Map<K, V> result = tryRemoveAll(keys);
    return (result != null) ? result : new HashMap<K, V>();
  }

  SimpleMap<K, V> $removeAll(List<K> keys) {
    return new SimpleMap.adapter(removeAll(keys));
  }

  Map<K, V> tryRemoveAll(List<K> keys) {
    MapSpliceResult<K, V> result = trySplice(keys, {});
    return (result != null) ? result.removedValues : null;
  }

  void removeValues(List<V> values) {
    Set<V> valueSet = values.toSet();
    Map<K, V> newValues = filter((V value) => !valueSet.contains(value));
    performSplice(newValues);
  }

  Map<K, V> clear() {
    Map<K, V> result = tryClear();
    return (result != null) ? result : new HashMap<K, V>();
  }

  SimpleMap<K, V> $clear() {
    return new SimpleMap.adapter(clear());
  }

  Map<K, V> tryClear() {
    Map<K, V> values = new MapUtil<K, V>().tryClear(_base);
    if ((values != null) && _ownsValues) {
      values.values.toList(growable: false).reversed.forEach(Destroyable.destroyObject);
    }
    return values;
  }

  MapSpliceResult<K, V> splice(List<K> removedKeys, Map<K, V> updatedValues) {
    MapSpliceResult<K, V> result = trySplice(removedKeys, updatedValues);
    return (result != null) ? result : new MapSpliceResult();
  }

  MapSpliceResult<K, V> trySplice(List<K> removedKeys, Map<K, V> updatedValues) {
    MapSpliceResult<K, V> spliceResult =
        new MapUtil<K, V>().trySplice(_base, removedKeys, updatedValues);
    if ((spliceResult != null) && _ownsValues) {
      spliceResult.removedValues.values.toList(growable: false).reversed.forEach(Destroyable.destroyObject);
    }
    return spliceResult;
  }

  Map<K, K> reindex(Map<K, K> keyMap) {
    Map<K, K> result = tryReindex(keyMap);
    return (result != null) ? result : new HashMap<K, K>();
  }

  Map<K, K> tryReindex(Map<K, K> keyMap) {
    return new MapUtil<K, V>().tryReindex(_base, keyMap);
  }

  MapSpliceParams<K, V> detectSplice(Map<K, V> newValues) {
    return new MapUtil<K, V>().detectSplice(_base, newValues);
  }

  Map<K, K> detectReindex(Map<K, V> newValues) {
    return new MapUtil<K, V>().detectReindex(_base, newValues);
  }

  void performSplice(Map<K, V> newValues) {
    MapSpliceParams<K, V> params = detectSplice(newValues);
    if (params != null) {
      trySplice(params.removedKeys, params.updatedValues);
    }
  }

  void performReindex(Map<K, V> newValues) {
    Map<K, K> keyMap = detectReindex(newValues);
    if (keyMap != null) {
      tryReindex(keyMap);
    }
  }

  AbstractMapMapper<K, V, dynamic> createMapper({dynamic create(V data),
    void destroy(dynamic item, V data)}) {
    return new AbstractMapMapper<K, V, dynamic>(this, create: create, destroy: destroy);
  }

  AbstractMapMapper<K, V, dynamic> createMapperTo({AbstractMap<K, dynamic> target,
    dynamic create(V data), void destroy(dynamic item, V data)}) {
    return new AbstractMapMapper<K, V, dynamic>(this, target: target, create: create, destroy: destroy);
  }

  AbstractMapFilterer<K, V> createFilterer({bool filter(V value)}) {
    return new AbstractMapFilterer<K, V>(this, filter: filter);
  }

  AbstractMapFilterer<K, V> createFiltererTo({AbstractMap<K, V> target, bool filter(V value)}) {
    return new AbstractMapFilterer<K, V>(this, target: target, filter: filter);
  }

  AbstractMapObserver<K, V> createObserver({void add(V value), void remove(V value),
    void clear(List<V> values), void change()}) {
    return new AbstractMapObserver<K, V>(this, add: add, remove: remove, clear: clear,
        change: change);
  }

  AbstractMapOrderer<K, V> createOrderer({AbstractList<V> target}) {
    return new AbstractMapOrderer<K, V>(this, target: target);
  }

  AbstractMapSorterComparing<K, V> createSorterComparing({AbstractList<V> target,
    int compare(V x, V y)}) {
    return new AbstractMapSorterComparing<K, V>(this, target: target, compare: compare);
  }

  AbstractMapIndexer<K, dynamic, V> createIndexer(
      {AbstractMap<dynamic, V> target, dynamic getKey(V value)}) {
    return new AbstractMapIndexer<K, dynamic, V>(this, target: target, getKey: getKey);
  }

  AbstractMapHeaper<K, V> createHeaper({AbstractSet<V> target}) {
    return new AbstractMapHeaper<K, V>(this, target: target);
  }

  AbstractMapInserter<K, V> createInserter({void add(V value, K key), void remove(V value, K key),
    void clear(Map<K, V> values)}) {
    return new AbstractMapInserter<K, V>(this, add: add, remove: remove, clear: clear);
  }

  bool equals(Map<K, V> other) {
    return new MapUtil<K, V>().equal(_base, other);
  }
}