part of insync;

class AbstractMapInserter<K, V> extends Owner {
  AbstractMap<K, V> _source;
  Function _addItem; // optional
  Function _removeItem; // optional
  Function _clearItems; // optional

  AbstractMap<K, V> get source => _source;

  AbstractMapInserter(this._source, {void add(V value, K key), void remove(V value, K key),
      void clear(Map<K, V> values)}) :
      _addItem = add,
      _removeItem = remove,
      _clearItems = clear {
    assert(_source != null);
    _add(_source.base);
  }

  void destroy() {
    _clear(_source.base);
    super.destroy();
  }

  void _add(Map<K, V> values) {
    if (_addItem == null) {
      return;
    }
    for (K key in values.keys) {
      _addItem(values[key], key);
    }
  }

  void _remove(Map<K, V> values) {
    if (_removeItem == null) {
      return;
    }
    for (K key in values.keys) {
      _removeItem(values[key], key);
    }
  }

  void _clear(Map<K, V> values) {
    if (values.isEmpty) {
      return;
    }
    if (_clearItems != null) {
      _clearItems(values);
    } else {
      _remove(values);
    }
  }
}