part of insync;

abstract class AbstractSet<V> extends AbstractCollection<V> {
  Set<V> _set = new Set();
  Set<V> _constantSet;

  AbstractSet<dynamic> createEmpty();

  AbstractSet() {
    _constantSet = new ConstantSet(_set);
  }

  AbstractSet.from(List<V> values) {
    _constantSet = new ConstantSet(_set);
    _set.addAll(values);
  }

  AbstractSet.adapter(this._set) {
    _constantSet = new ConstantSet(_set);
  }

  Set<V> get set => _constantSet;
  int get length => _set.length;
  bool get isEmpty => _set.isEmpty;
  V get first => _set.isEmpty ? null : _set.first;

  bool containsValue(V value) {
    return _set.contains(value);
  }

  bool contains(V value) => containsValue(value);

  bool every(bool f(V value)) {
    for (V value in _set) {
      if (!f(value)) {
        return false;
      }
    }
    return true;
  }

  Set<V> filter(bool f(V value)) {
    Set<V> result = new Set();
    every((V v) { if (f(v) != false) { result.add(v); }});
    return result;
  }

  SimpleSet<V> $filter(bool f(V value)) {
    return new SimpleSet.adapter(filter(f));
  }

  Set<dynamic> map(dynamic f(V value)) {
    Set<dynamic> result = new Set();
    every((V v) { result.add(f(v)); });
    return result;
  }

  SimpleSet<dynamic> $map(dynamic f(V value)) {
    return new SimpleSet.adapter(map(f));
  }

  Set<V> asSet() {
    return set;
  }

  AbstractSet<V> $asSet() {
    return this;
  }

  bool add(V value) {
    return tryAdd(value) != null;
  }

  bool tryAdd(V value) {
    return (trySplice([], [value]) != null) ? true : null;
  }

  List<V> addAll(List<V> values) {
    List<V> result = tryAddAll(values);
    return (result != null) ? result : new List();
  }

  SimpleList<V> $addAll(List<V> values) {
    return new SimpleList.adapter(addAll(values));
  }

  List<V> tryAddAll(List<V> values) {
    SetSpliceResult<V> result = trySplice([], values);
    return (result != null) ? result.addedValues : null;
  }

  bool remove(V value) {
    return tryRemove(value) != null;
  }

  bool tryRemove(V value) {
    return (trySplice([value], []) != null) ? true : null;
  }

  void removeValue(V value) {
    tryRemove(value);
  }

  List<V> removeAll(List<V> values) {
    List<V> result = tryRemoveAll(values);
    return (result != null) ? result : new List();
  }

  SimpleList<V> $removeAll(List<V> values) {
    return new SimpleList.adapter(removeAll(values));
  }

  List<V> tryRemoveAll(List<V> values) {
    SetSpliceResult<V> result = trySplice(values, []);
    return (result != null) ? result.removedValues : null;
  }

  void removeValues(List<V> values) {
    tryRemoveAll(values);
  }

  List<V> clear() {
    List<V> result = tryClear();
    return (result != null) ? result : new List();
  }

  SimpleList<V> $clear() {
    return new SimpleList.adapter(clear());
  }

  List<V> tryClear() {
    List<V> values = new SetUtil<V>().tryClear(_set);
    if ((values != null) && _ownsValues) {
      values.reversed.forEach(Destroyable.destroyObject);
    }
    return values;
  }

  SetSpliceResult<V> splice(List<V> removedValues, List<V> addedValues) {
    SetSpliceResult<V> result = trySplice(removedValues, addedValues);
    return (result != null) ? result : new SetSpliceResult();
  }

  SetSpliceResult<V> trySplice(List<V> removedValues, List<V> addedValues) {
    SetSpliceResult<V> spliceResult = new SetUtil<V>().trySplice(_set, removedValues, addedValues);
    if ((spliceResult != null) && _ownsValues) {
      spliceResult.removedValues.reversed.forEach(Destroyable.destroyObject);
    }
    return spliceResult;
  }

  SetSpliceParams<V> detectSplice(List<V> newValues) {
    return new SetUtil<V>().detectSplice(_set, newValues);
  }

  void performSplice(List<V> newValues) {
    SetSpliceParams<V> spliceParams = detectSplice(newValues);
    if (spliceParams != null) {
      trySplice(spliceParams.removedValues, spliceParams.addedValues);
    }
  }

  AbstractSetMapper<V, dynamic> createMapper({dynamic create(V data),
    void destroy(dynamic item, V data)}) {
    return new AbstractSetMapper<V, dynamic>(this, create: create, destroy: destroy);
  }

  AbstractSetMapper<V, dynamic> createMapperTo({AbstractSet<dynamic> target,
    dynamic create(V data), void destroy(dynamic item, V data)}) {
    return new AbstractSetMapper<V, dynamic>(this, target: target, create: create, destroy: destroy);
  }

  AbstractSetFilterer<V> createFilterer({bool filter(V value)}) {
    return new AbstractSetFilterer<V>(this, filter: filter);
  }

  AbstractSetFilterer<V> createFiltererTo({AbstractSet<V> target, bool filter(V value)}) {
    return new AbstractSetFilterer<V>(this, target: target, filter: filter);
  }

  AbstractSetObserver<V> createObserver({void add(V value), void remove(V value),
    void clear(List<V> values), void change()}) {
    return new AbstractSetObserver<V>(this, add: add, remove: remove, clear: clear,
        change: change);
  }

  AbstractSetOrderer<V> createOrderer({AbstractList<V> target}) {
    return new AbstractSetOrderer<V>(this, target: target);
  }

  AbstractSetSorterComparing<V> createSorterComparing({AbstractList<V> target,
    int compare(V x, V y)}) {
    return new AbstractSetSorterComparing<V>(this, target: target, compare: compare);
  }

  AbstractSetIndexer<dynamic, V> createIndexer(
      {AbstractMap<dynamic, V> target, dynamic getKey(V value)}) {
    return new AbstractSetIndexer<dynamic, V>(this, target: target, getKey: getKey);
  }

  AbstractSetHeaper<V> createHeaper({AbstractSet<V> target}) {
    return new AbstractSetHeaper<V>(this, target: target);
  }

  bool equals(List<V> values) {
    return new SetUtil<V>().equal(_set, values);
  }
}