part of insync;

abstract class IndexedCollection<K, V> extends AbstractCollection<V> {
  K get firstKey;
  V operator [](K key);
  List<K> getKeys();
  IndexedCollection<K, V> $clear();

  bool every2(bool f(V value, K key));
  IndexedCollection<K, V> $filter(bool f(V value));
  dynamic filter2(bool f(V value, K key));
  IndexedCollection<K, V> $filter2(bool f(V value, K key));
  IndexedCollection<K, dynamic> $map(bool f(V value));
  dynamic map2(dynamic f(V value, K key));
  IndexedCollection<K, dynamic> $map2(bool f(V value, K key));
  IndexedCollection<K, dynamic> createEmpty();

  Wrap<V> trySet(V value, K key);
  V tryRemove(K key);

  V get first {
    K key = firstKey;
    return (key != null) ? this[key] : null;
  }

  SimpleList<K> $getKeys() {
    return new SimpleList.adapter(getKeys());
  }

  bool containsKey(K key) {
    return this[key] != null;
  }

  bool containsValue(V value) {
    return !every((V v) => (value != v));
  }

  K keyOf(V value) {
    return find2((V v, K k) => (value == v));
  }

  void operator []=(K key, V value) {
    trySet(value, key);
  }

  V set(V value, K key) {
    Wrap<V> result = trySet(value, key);
    return (result != null) ? result.value : this[key];
  }

  V remove(K key) {
    return tryRemove(key);
  }

  K removeValue(V value) {
    K key = keyOf(value);
    if (key != null) {
      tryRemove(key);
    }
    return key;
  }

  bool every(bool f(V value)) {
    return every2((V v, K k) => f(v));
  }

  K find(bool f(V value)) {
    return find2((V v, K k) => f(v));
  }

  bool some2(bool f(V value, K key)) {
    return !every2((V v, K k) => (f(v, k) == false));
  }

  void each2(void f(V value, K key)) {
    every2((V v, K k) { f(v, k); return true; });
  }

  K find2(bool f(V value, K key)) {
    K result;
    every2((V v, K k) { if (f(v, k) != false) { result = k; return false; } return true; });
    return result;
  }

  V search2(bool f(V value, K key)) {
    V result;
    every2((V v, K k) { if (f(v, k) != false) { result = v; return false; } return true; });
    return result;
  }

  List<K> getSortingKeys([Comparable callback(V value, K key), int order = 1]) {
    if (callback == null) {
      callback = (V v, K k) => v;
    }
    List<Pair<K, Comparable>> pairs = new List();
    every2((V v, K k) { pairs.add(new Pair(k, callback(v, k))); });
    pairs.sort((Pair<K, Comparable> x, Pair<K, Comparable> y) => order * x.y.compareTo(y.y));
    return pairs.map((Pair<K, Comparable> t) => t.x).toList(growable: true);
  }

  SimpleList<K> $getSortingKeys([Comparable callback(V value, K key), int order = 1]) {
    return new SimpleList.adapter(getSortingKeys(callback, order));
  }

  List<K> getSortingKeysComparing(int compare(V v1, V v2, K k1, K k2), [int order = 1]) {
    List<Pair<K, V>> pairs = new List();
    every2((V v, K k) { pairs.add(new Pair(k, v)); });
    pairs.sort((Pair<K, V> x, Pair<K, V> y) => order * compare(x.y, y.y, x.x, y.x));
    return pairs.map((Pair<K, Comparable> t) => t.x).toList(growable: true);
  }

  SimpleList<K> $getSortingKeysComparing(int compare(V v1, V v2, K k1, K k2), [int order = 1]) {
    return new SimpleList.adapter(getSortingKeysComparing(compare, order));
  }

  List<V> toSorted2([Comparable callback(V value, K key), int order = 1]) {
    return getSortingKeys(callback, order).map((K k) => this[k]).toList(growable: true);
  }

  SimpleList<V> $toSorted2([Comparable callback(V value, K key), int order = 1]) {
    return new SimpleList.adapter(toSorted2(callback, order));
  }

  List<V> toSortedComparing2(int compare(V v1, V v2, K k1, K k2), [int order = 1]) {
    return getSortingKeysComparing(compare, order).map((K k) => this[k]).toList(growable: true);
  }

  SimpleList<V> $toSortedComparing2(int compare(V v1, V v2, K k1, K k2), [int order = 1]) {
    return new SimpleList.adapter(toSortedComparing2(compare, order));
  }

  Map<dynamic, V> index2(dynamic f(V value, K key)) {
    Map<dynamic, V> result = new HashMap();
    every2((V v, K k) { dynamic key = f(v, k); if (key != null) { result[key] = v; } return true; });
    return result;
  }

  SimpleMap<dynamic, V> $index2(dynamic f(V value, K key)) {
    return new SimpleMap.adapter(index2(f));
  }

  Map<K, V> toMap() {
    Map<K, V> result = new HashMap();
    every2((V v, K k) { result[k] = v; return true; });
    return result;
  }

  SimpleMap<K, V> $toMap() {
    return new SimpleMap.adapter(toMap());
  }

  Map<K, V> asMap() {
    return toMap();
  }

  AbstractMap<K, V> $asMap() {
    return $toMap();
  }
}