part of insync;

class ObservableList<V> extends AbstractList<V> implements ObservableCollection<V> {
  EventDispatcher<ObservableListSpliceEventParams<V>> onSplice = new EventDispatcher();
  EventDispatcher<ObservableListReplaceEventParams<V>> onReplace = new EventDispatcher();
  EventDispatcher<ObservableListMoveEventParams<V>> onMove = new EventDispatcher();
  EventDispatcher<List<V>> onClear = new EventDispatcher();
  EventDispatcher<ObservableListReorderEventParams> onReorder = new EventDispatcher();
  EventDispatcher onChange = new EventDispatcher();

  Property<int> _length;
  Property<int> get lengthProperty => _length;

  ObservableList() : super(), _length = new Property(0);

  ObservableList.from(List<V> values) : super.from(values), _length = new Property(values.length);

  ObservableList.adapter(List<V> list) : super.adapter(list), _length = new Property(list.length);

  ObservableList.filled(int length, V value) : super.filled(length, value),
    _length = new Property(length);

  void destroy() {
    onSplice.destroy();
    onReplace.destroy();
    onMove.destroy();
    onClear.destroy();
    onReorder.destroy();
    onChange.destroy();
    _length.destroy();
    super.destroy();
  }

  Wrap<V> trySet(V value, int index) {
    Wrap<V> wrap = super.trySet(value, index);
    if (wrap == null) {
      return null;
    }
    onReplace.fire(new ObservableListReplaceEventParams<V>(index, wrap.value, value));
    onChange.fire();
    return wrap;
  }

  V tryMove(int fromIndex, int toIndex) {
    V value = super.tryMove(fromIndex, toIndex);
    if (value == null) {
      return null;
    }
    onMove.fire(new ObservableListMoveEventParams<V>(value, fromIndex, toIndex));
    onChange.fire();
    return value;
  }

  List<V> tryClear() {
    List<V> oldValues = super.tryClear();
    if (oldValues == null) {
      return null;
    }
    _length.value = 0;
    onClear.fire(oldValues);
    onChange.fire();
    return oldValues;
  }

  ListSpliceResult<V> trySplice(List<IndexCount> removeParamsList, List<IndexValues<V>> addParamsList) {
    ListSpliceResult<V> spliceResult = super.trySplice(removeParamsList, addParamsList);
    if (spliceResult == null) {
      return null;
    }
    _length.value = length;
    onSplice.fire(new ObservableListSpliceEventParams<V>(spliceResult));
    onChange.fire();
    return spliceResult;
  }

  List<V> tryReorder(List<int> indexList) {
    List<V> oldValues = super.tryReorder(indexList);
    if (oldValues == null) {
      return null;
    }
    onReorder.fire(new ObservableListReorderEventParams(indexList, oldValues));
    onChange.fire();
    return oldValues;
  }

  ObservableList<dynamic> createEmpty() {
    return new ObservableList<V>();
  }

  ObservableList<dynamic> createEmptyList() {
    return new ObservableList();
  }

  ObservableMap<dynamic, dynamic> createEmptyMap() {
    return new ObservableMap();
  }

  ObservableSet<dynamic> createEmptySet() {
    return new ObservableSet();
  }

  ObservableListMapper<V, dynamic> createMapper({dynamic create(V data),
    void destroy(dynamic item, V data)}) {
    return new ObservableListMapper(this, create: create, destroy: destroy);
  }

  ObservableListMapper<V, dynamic> createMapperTo({AbstractList<dynamic> target,
    dynamic create(V data), void destroy(dynamic item, V data)}) {
    return new ObservableListMapper(this, target: target, create: create, destroy: destroy);
  }

  ObservableListFilterer<V> createFilterer({bool filter(V value)}) {
    return new ObservableListFilterer(this, filter: filter);
  }

  ObservableListFilterer<V> createFiltererTo({AbstractList<dynamic> target,
    bool filter(V value)}) {
    return new ObservableListFilterer(this, target: target, filter: filter);
  }

  ObservableListObserver<V> createObserver({void add(V value), void remove(V value),
    void clear(List<V> values), void change()}) {
    return new ObservableListObserver<V>(this, add: add, remove: remove, clear: clear,
        change: change);
  }

  ObservableListOrderer<V> createOrderer({AbstractList<V> target}) {
    return new ObservableListOrderer<V>(this, target: target);
  }

  ObservableListSorterComparing<V> createSorterComparing({AbstractList<V> target,
    int compare(V x, V y)}) {
    return new ObservableListSorterComparing<V>(this, target: target, compare: compare);
  }

  ObservableListIndexer<dynamic, V> createIndexer(
      {AbstractMap<dynamic, V> target, dynamic getKey(V value)}) {
    return new ObservableListIndexer<dynamic, V>(this, target: target, getKey: getKey);
  }

  ObservableListHeaper<V> createHeaper({AbstractSet<V> target}) {
    return new ObservableListHeaper<V>(this, target: target);
  }

  ObservableListInserter<V> createInserter({void add(V value, int index),
    void remove(V value, int index), void clear(List<V> values)}) {
    return new ObservableListInserter<V>(this, add: add, remove: remove, clear: clear);
  }

  ObservableListMerger<dynamic> createMerger({AbstractList<dynamic> target}) {
    return new ObservableListMerger<dynamic>(this, target: target);
  }

  _ObservableListMergerBunch<dynamic> _createMergerBunch(ObservableListMerger<dynamic> merger) {
    return new _ObservableListMergerBunch<dynamic>(merger, this);
  }

  ObservableListReverser<V> createReverser({AbstractList<dynamic> target}) {
    return new ObservableListReverser<V>(this, target: target);
  }
}

class ObservableListSpliceEventParams<V> {
  final ListSpliceResult<V> spliceResult;
  const ObservableListSpliceEventParams(this.spliceResult);
}

class ObservableListMoveEventParams<V> {
  final int fromIndex;
  final int toIndex;
  final V value;
  const ObservableListMoveEventParams(this.fromIndex, this.toIndex, this.value);
}

class ObservableListReplaceEventParams<V> {
  final int index;
  final V oldValue;
  final V newValue;
  const ObservableListReplaceEventParams(this.index, this.oldValue, this.newValue);
}

class ObservableListValuesEventParams<V> {
  final List<V> values;
  SimpleList<V> get $values => new SimpleList.adapter(values);
  const ObservableListValuesEventParams(this.values);
}

class ObservableListReorderEventParams<V> extends ObservableListValuesEventParams<V> {
  final List<int> indexList;
  SimpleList<int> get $indexList => new SimpleList.adapter(indexList);
  const ObservableListReorderEventParams(this.indexList, List<V> values) : super(values);
}