part of insync;

class ObservableListFilterer<V> extends AbstractListFilterer<V> {
  ObservableListFilterer(ObservableList<V> source, {AbstractList<V> target,
    bool filter(V value)}) : super(source, target: target, filter: filter) {
    own(source.onSplice.listen(_onSplice));
    own(source.onReplace.listen(_onReplace));
    own(source.onMove.listen(_onMove));
    own(source.onClear.listen(_onClear));
    own(source.onReorder.listen(_onReorder));
  }

  void _onSplice(ObservableListSpliceEventParams<V> params) {
    ListSpliceResult<V> spliceResult = params.spliceResult;
    _splice(spliceResult.removedValuesList, spliceResult.addedValuesList);
  }

  void _onReplace(ObservableListReplaceEventParams<V> params) {
    bool oldFiltered = _filtered[params.index] != 0;
    bool newFiltered = _filterValue(params.newValue) != false;
    if (!oldFiltered && !newFiltered) {
      return;
    }
    int index = _countFiltered(0, params.index);
    _filtered[params.index] = newFiltered ? 1 : 0;
    if (!newFiltered) {
      _target.tryRemove(index);
    } else if (!oldFiltered) {
      _target.tryAdd(params.newValue, index);
    } else {
      _target.trySet(params.newValue, index);
    }
  }

  void _onMove(ObservableListMoveEventParams<V> params) {
    if (_filtered[params.fromIndex] != 0) {
      int fromIndex, toIndex;
      if (params.fromIndex < params.toIndex) {
        fromIndex = _countFiltered(0, params.fromIndex);
        toIndex = fromIndex + _countFiltered(params.fromIndex + 1, params.toIndex - params.fromIndex);
      } else {
        toIndex = _countFiltered(0, params.toIndex);
        fromIndex = toIndex + _countFiltered(params.toIndex, params.fromIndex - params.toIndex);
      }
      _target.tryMove(fromIndex, toIndex);
    }
    new ListUtil<V>().tryMove(_filtered, params.fromIndex, params.toIndex);
  }

  void _onClear(List<V> params) {
    _target.tryClear();
  }

  void _onReorder(ObservableListReorderEventParams<V> params) {
    int targetIndex = 0;
    Map<int, int> targetIndexWhichMovesToI = new Map<int, int>();
    for (int sourceIndex = 0, l = _filtered.length; sourceIndex < l; ++sourceIndex) {
      if (_filtered[sourceIndex] != 0) {
        targetIndexWhichMovesToI[params.indexList[sourceIndex]] = targetIndex++;
      }
    }
    new ListUtil<V>().tryReorder(_filtered, params.indexList);

    targetIndex = 0;
    List<int> indexes = new List<int>(_target.length);
    for (int sourceIndex = 0, l = _filtered.length; sourceIndex < l; ++sourceIndex) {
      if (_filtered[sourceIndex] != 0) {
        indexes[targetIndexWhichMovesToI[sourceIndex]] = targetIndex++;
      }
    }
    _target.tryReorder(indexes);
  }
}