part of insync;

class ObservableListInserter<V> extends AbstractListInserter<V> {
  ObservableListInserter(ObservableList<V> source, {void add(V value, int index),
    void remove(V value, int index), void clear(List<V> values)}) :
      super(source, add: add, remove: remove, clear: clear) {
    own(source.onSplice.listen(_onSplice));
    own(source.onReplace.listen(_onReplace));
    own(source.onMove.listen(_onMove));
    own(source.onClear.listen(_onClear));
    own(source.onReorder.listen(_onReorder));
  }

  void _onSplice(ObservableListSpliceEventParams<V> params) {
    ListSpliceResult<V> spliceResult = params.spliceResult;
    List<V> oldValues = spliceResult.oldValues;
    List<V> removedValues = spliceResult.removedValues;

    // if there is an effective clearing function, just reset the controller
    if ((_clearValues != null) && (3 * removedValues.length > 2 * oldValues.length)) {
      _clearValues(oldValues);
      _add(_source.list, 0);
      return;
    }

    // else, splice the elements
    List<IndexValues<V>> removedValuesList = spliceResult.removedValuesList;
    List<IndexValues<V>> addedValuesList = spliceResult.addedValuesList;
    for (int i = removedValuesList.length - 1; i >= 0; --i) {
      IndexValues<V> removeParams = removedValuesList[i];
      _remove(removeParams.values, removeParams.index);
    }
    for (int i = 0, l = addedValuesList.length; i < l; ++i) {
      IndexValues<V> addParams = addedValuesList[i];
      _add(addParams.values, addParams.index);
    }
  }

  void _onReplace(ObservableListReplaceEventParams<V> params) {
    if (_removeValue != null) {
      _removeValue(params.oldValue, params.index);
    }
    if (_addValue != null) {
      _addValue(params.newValue, params.index);
    }
  }

  void _onMove(ObservableListMoveEventParams<V> params) {
    if (_removeValue != null) {
      _removeValue(params.value, params.fromIndex);
    }
    if (_addValue != null) {
      _addValue(params.value, params.toIndex);
    }
  }

  void _onClear(List<V> values) {
    _clear(values);
  }

  void _onReorder(ObservableListReorderEventParams<V> params) {
    _clear(params.values);
    _add(_source.list, 0);
  }
}