part of insync;

class ObservableListMapper<S, T> extends AbstractListMapper<S, T> {
  ObservableListMapper(ObservableList<S> source,
      {AbstractList<T> target, T create(S data), void destroy(T item, S data)}) :
      super(source, target: target, create: create, destroy: destroy) {
    own(source.onSplice.listen(_onSplice));
    own(source.onReplace.listen(_onReplace));
    own(source.onMove.listen(_onMove));
    own(source.onClear.listen(_onClear));
    own(source.onReorder.listen(_onReorder));
  }

  void _onSplice(ObservableListSpliceEventParams<S> params) {
    ListSpliceResult<S> sourceResult = params.spliceResult;
    List<IndexValues<S>> sourceAddedValuesList = sourceResult.addedValuesList;
    List<IndexValues<T>> targetAddParamsList = new List();
    for (int i = 0; i < sourceAddedValuesList.length; ++i) {
      IndexValues<S> addedValues = sourceAddedValuesList[i];
      targetAddParamsList.add(new IndexValues(addedValues.index, _create(addedValues.values)));
    }
    ListSpliceResult<T> targetResult = _target.trySplice(sourceResult.removeParamsList, targetAddParamsList);
    List<IndexValues<S>> sourceRemovedValuesList = sourceResult.removedValuesList;
    List<IndexValues<T>> targetRemovedValuesList = targetResult.removedValuesList;
    for (int i = targetRemovedValuesList.length - 1; i >= 0; --i) {
      _destroy(targetRemovedValuesList[i].values, sourceRemovedValuesList[i].values);
    }
  }

  void _onReplace(ObservableListReplaceEventParams<S> params) {
    T newValue = _createItem(params.newValue);
    T oldValue = _target.trySet(newValue, params.index);
    if (_destroyItem != null) {
      _destroyItem(oldValue, params.oldValue);
    }
  }

  void _onMove(ObservableListMoveEventParams<S> params) {
    _target.tryMove(params.fromIndex, params.toIndex);
  }

  void _onClear(List<S> values) {
    _destroy(target.tryClear(), values);
  }

  void _onReorder(ObservableListReorderEventParams<S> params) {
    _target.tryReorder(params.indexList);
  }
}