part of insync;

class ObservableListObserver<V> extends AbstractListObserver<V> {
  ObservableListObserver(ObservableList<V> source, {void add(V value), void remove(V value),
    void clear(List<V> values), void change()}) : super(source, add: add, remove: remove,
        clear: clear, change: change) {
    own(source.onSplice.listen(_onSplice));
    own(source.onReplace.listen(_onReplace));
    own(source.onClear.listen(_onClear));
    if (_change != null) {
      own(source.onChange.listen((p) { _change(); }));
    }
  }

  void _onSplice(ObservableListSpliceEventParams<V> params) {
    ListSpliceResult<V> spliceResult = params.spliceResult;
    List<V> oldValues = spliceResult.oldValues;
    List<V> removedValues = spliceResult.removedValues;

    if ((_clearValues != null) && (3 * removedValues.length > 2 * oldValues.length)) {
      // if there is an effective clearing function, just reset the controller
      _clear(oldValues);
      _add(_source.list);
    } else {
      // else, splice the elements
      _remove(removedValues);
      _add(spliceResult.addedValues);
    }
  }

  void _onReplace(ObservableListReplaceEventParams<V> params) {
    if (_removeValue != null) {
      _removeValue(params.oldValue);
    }
    if (_addValue != null) {
      _addValue(params.newValue);
    }
  }

  void _onClear(List<V> values) {
    _clear(values);
  }
}