part of insync;

class ObservableMap<K, V> extends AbstractMap<K, V> implements ObservableCollection<V> {
  EventDispatcher<MapSpliceResult<K, V>> onSplice = new EventDispatcher();
  EventDispatcher<Map<K, K>> onReindex = new EventDispatcher();
  EventDispatcher<Map<K, V>> onClear = new EventDispatcher();
  EventDispatcher onChange = new EventDispatcher();

  Property<int> _length;
  Property<int> get lengthProperty => _length;

  ObservableMap() : super(), _length = new Property<int>(0);

  ObservableMap.from(Map<K, V> values) :
    super.from(values), _length = new Property<int>(values.length);

  ObservableMap.adapter(Map<K, V> map) :
    super.adapter(map), _length = new Property<int>(map.length);

  void destroy() {
    onSplice.destroy();
    onReindex.destroy();
    onClear.destroy();
    onChange.destroy();
    _length.destroy();
    super.destroy();
  }

  MapSpliceResult<K, V> trySplice(List<K> removedKeys, Map<K, V> updatedValues) {
    MapSpliceResult<K, V> spliceResult = super.trySplice(removedKeys, updatedValues);
    if (spliceResult == null) {
      return null;
    }
    _length.value = length;
    onSplice.fire(spliceResult);
    onChange.fire();
    return spliceResult;
  }

  Map<K, V> tryClear() {
    Map<K, V> values = super.tryClear();
    if (values == null) {
      return null;
    }
    _length.value = 0;
    onClear.fire(values);
    onChange.fire();
    return values;
  }

  Map<K, K> tryReindex(Map<K, K> keyMap) {
    Map<K, K> result = super.tryReindex(keyMap);
    if (result == null) {
      return null;
    }
    onReindex.fire(result);
    onChange.fire();
    return result;
  }

  ObservableMap<K, dynamic> createEmpty() {
    return new ObservableMap<K, V>();
  }

  ObservableList<dynamic> createEmptyList() {
    return new ObservableList();
  }

  ObservableMap<dynamic, dynamic> createEmptyMap() {
    return new ObservableMap();
  }

  ObservableSet<dynamic> createEmptySet() {
    return new ObservableSet();
  }

  ObservableMapMapper<K, V, dynamic> createMapper({dynamic create(V data),
    void destroy(dynamic item, V data)}) {
    return new ObservableMapMapper(this, create: create, destroy: destroy);
  }

  ObservableMapMapper<K, V, dynamic> createMapperTo({AbstractMap<K, dynamic> target,
    dynamic create(V data), void destroy(dynamic item, V data)}) {
    return new ObservableMapMapper(this, target: target, create: create, destroy: destroy);
  }

  ObservableMapFilterer<K, V> createFilterer({bool filter(V value)}) {
    return new ObservableMapFilterer<K, V>(this, filter: filter);
  }

  ObservableMapFilterer<K, V> createFiltererTo({AbstractMap<K, V> target, bool filter(V value)}) {
    return new ObservableMapFilterer<K, V>(this, target: target, filter: filter);
  }

  ObservableMapObserver<K, V> createObserver({void add(V value), void remove(V value),
    void clear(List<V> values), void change()}) {
    return new ObservableMapObserver<K, V>(this, add: add, remove: remove, clear: clear,
        change: change);
  }

  ObservableMapOrderer<K, V> createOrderer({AbstractList<V> target}) {
    return new ObservableMapOrderer<K, V>(this, target: target);
  }

  ObservableMapSorterComparing<K, V> createSorterComparing({AbstractList<V> target,
    int compare(V x, V y)}) {
    return new ObservableMapSorterComparing<K, V>(this, target: target, compare: compare);
  }

  ObservableMapIndexer<K, dynamic, V> createIndexer(
      {AbstractMap<dynamic, V> target, dynamic getKey(V value)}) {
    return new ObservableMapIndexer<K, dynamic, V>(this, target: target, getKey: getKey);
  }

  ObservableMapHeaper<K, V> createHeaper({AbstractSet<V> target}) {
    return new ObservableMapHeaper<K, V>(this, target: target);
  }

  ObservableMapInserter<K, V> createInserter({void add(V value, K key),
    void remove(V value, K key), void clear(Map<K, V> values)}) {
    return new ObservableMapInserter<K, V>(this, add: add, remove: remove, clear: clear);
  }
}