part of insync;

class SetSpliceParams<V> {
  List<V> removedValues;
  SimpleList<V> get $removedValues => new SimpleList.adapter(removedValues);

  List<V> addedValues;
  SimpleList<V> get $addedValues => new SimpleList.adapter(addedValues);

  SetSpliceParams() : removedValues = new List(), addedValues = new List() {}

  bool get isEmpty {
    return removedValues.isEmpty && addedValues.isEmpty;
  }
}

class SetSpliceResult<V> {
  List<V> removedValues;
  SimpleList<V> get $removedValues => new SimpleList.adapter(removedValues);

  List<V> addedValues;
  SimpleList<V> get $addedValues => new SimpleList.adapter(addedValues);

  SetSpliceResult() : removedValues = new List(), addedValues = new List() {}
  SetSpliceResult.from(this.removedValues, this.addedValues) {}

  bool get isEmpty {
    return removedValues.isEmpty && addedValues.isEmpty;
  }
}

class SetUtil<V> {
  bool tryAdd(Set<V> target, V value) {
    if (target.contains(value)) {
      return null;
    }
    target.add(value);
    return true;
  }

  List<V> tryAddAll(Set<V> target, List<V> values) {
    List<V> addedValues = new List<V>();
    for (int i = 0; i < values.length; ++i) {
      V value = values[i];
      if (tryAdd(target, value) != null) {
        addedValues.add(value);
      }
    }
    return addedValues.isEmpty ? null : addedValues;
  }

  bool tryRemove(Set<V> target, V value) {
    if (!target.contains(value)) {
      return null;
    }
    target.remove(value);
    return true;
  }

  List<V> tryRemoveAll(Set<V> target, List<V> values) {
    List<V> removedValues = new List<V>();
    for (int i = 0; i < values.length; ++i) {
      V value = values[i];
      if (tryRemove(target, value) != null) {
        removedValues.add(value);
      }
    }
    return removedValues.isEmpty ? null : removedValues;
  }

  List<V> tryClear(Set<V> target) {
    if (target.isEmpty) {
      return null;
    }
    List<V> values = target.toList(growable: true);
    target.clear();
    return values;
  }

  SetSpliceResult<V> trySplice(Set<V> target, List<V> removedValues, List<V> addedValues) {
    SetSpliceResult<V> spliceResult = new SetSpliceResult.from(
        tryRemoveAll(target, removedValues),
        tryAddAll(target, addedValues));
    if (spliceResult.removedValues == null) {
      spliceResult.removedValues = new List();
    }
    if (spliceResult.addedValues == null) {
      spliceResult.addedValues = new List();
    }
    return spliceResult.isEmpty ? null : spliceResult;
  }

  SetSpliceParams<V> detectSplice(Set<V> oldValues, List<V> newValueList) {
    SetSpliceParams<V> spliceParams = new SetSpliceParams();
    Set<V> newValues = newValueList.toSet();
    for (V value in oldValues) {
      if (!newValues.contains(value)) {
        spliceParams.removedValues.add(value);
      }
    }
    for (V value in newValues) {
      if (!oldValues.contains(value)) {
        spliceParams.addedValues.add(value);
      }
    }
    return spliceParams.isEmpty ? null : spliceParams;
  }

  void toggle(Set<V> set, V value, bool contains) {
    contains ? set.add(value) : set.remove(value);
  }

  bool equal(Set<V> x, List<V> y) {
    return (x.length == y.length) && x.containsAll(y);
  }
}