part of insync;

class PropertySwitcher<T> extends Owner {
  final List<Property> sources;
  final Function _createValue;
  final Function _destroyValue;
  final bool _acceptNull;
  List _sourceValues;
  T _targetValue;

  Property<T> _target;
  Property<T> get target => _target;

  PropertySwitcher(this.sources,
      {Function create, Function destroy, Property<T> target, bool acceptNull}) :
      _createValue = create, _destroyValue = destroy, _acceptNull = acceptNull {
    _target = (target != null) ? target : new Property<T>()..ownBy(this);
    _init();
    sources.forEach(watch);
  }

  void destroy() {
    _done();
    super.destroy();
  }

  void listen(EventDispatcher dispatcher) {
    own(dispatcher.listen(_update));
  }

  void watch(Property property) {
    listen(property.onChange);
  }

  void _update([dynamic p]) {
    _done();
    _init();
  }

  void _init() {
    List values = sources.map((Property source) => source.value).toList();
    _sourceValues = (_acceptNull || values.every((value) => value != null)) ? values : null;
    if (_sourceValues != null) {
      _targetValue = Function.apply(_createValue, _sourceValues);
      _target.value = _targetValue;
    }
  }

  void _done() {
    if (_target.value == _targetValue) {
      _target.value = null;
    }
    if ((_sourceValues != null) && (_destroyValue != null)) {
      List args = [_targetValue];
      args.addAll(_sourceValues);
      Function.apply(_destroyValue, args);
    }
    _sourceValues = null;
    _targetValue = null;
  }
}